from typing import Literal, Optional, Tuple, TypedDict

import numpy as np
from numpy import linalg
from scipy.interpolate import interp1d


class DifferentialPathlengthFactors(TypedDict):
    baby_head: float
    adult_head: float
    adult_arm: float
    adult_leg: float


DpfType = Literal["baby_head", "adult_head", "adult_arm", "adult_leg"]


class UCLNConstants:
    """Class for holding MBL constants"""

    # Differential path length factors based on Dunan 1994
    _dpf_dict: DifferentialPathlengthFactors = {
        "baby_head": 4.99,
        "adult_head": 6.26,
        "adult_arm": 4.16,
        "adult_leg": 5.51,
    }

    def __init__(
        self,
        extinction_coefficients: np.ndarray,
        wavelength_dependency_of_pathlength: np.ndarray,
        optode_dist: float,
        dpf_type: DpfType,
        wavelengths: Tuple[float, float],
    ) -> None:
        self.extinction_coefficients = extinction_coefficients
        self.wavelength_dependency = wavelength_dependency_of_pathlength
        self.optode_dist = optode_dist
        self.dpf: float = self._dpf_dict[dpf_type]

        (min_wavelength, max_wavelength) = wavelengths
        self.interp_wavelengths: np.ndarray = np.arange(
            min_wavelength, max_wavelength + 1
        )


class UCLN:
    """Class for calculating conc. using the Modified Beer-Lambert Law"""

    def __init__(self, constants: UCLNConstants) -> None:
        self.constants: UCLNConstants = constants
        self.attenuation_interp_wavelength_dependency: Optional[
            np.ndarray
        ] = None

    def _calc_change_in_attenuation(
        self, spectra: np.ndarray, spectra_wavelengths: np.ndarray
    ) -> None:
        # Preallocate arrays
        n_spectra = spectra.shape[0]
        attenuation: np.ndarray = np.zeros(spectra.shape)
        attenuation_interp: np.ndarray = np.zeros(
            (self.constants.interp_wavelengths.size, n_spectra)
        )

        for i in range(n_spectra):
            attenuation[i, :] = np.log10(spectra[0, :] / spectra[i, :])
            attenuation_interp[:, i] = interp1d(
                spectra_wavelengths,
                attenuation[i, :].T,
                kind="cubic",
            )(self.constants.interp_wavelengths)

        self.attenuation_interp_wavelength_dependency = np.divide(
            attenuation_interp.T, self.constants.wavelength_dependency
        )

    def calc_concentrations(
        self, spectra: np.ndarray, spectra_wavelengths: np.ndarray
    ) -> Optional[np.ndarray]:
        # Calculate change in attenuation
        self._calc_change_in_attenuation(spectra, spectra_wavelengths)

        # Note: The inverse of the matrix isn't unique meaning these differ
        # from the MATLAB equivalents
        ext_coeffs_inv: np.ndarray = linalg.pinv(
            self.constants.extinction_coefficients
        )

        optode_dist = self.constants.optode_dist
        dpf = self.constants.dpf

        if self.attenuation_interp_wavelength_dependency is not None:
            return np.transpose(
                np.matmul(
                    ext_coeffs_inv,
                    self.attenuation_interp_wavelength_dependency.T,
                )
                * (1 / (optode_dist * dpf))
            )
        else:
            return None



from enum import Enum, auto

import numpy as np
from scipy.optimize import OptimizeResult, minimize

# ADapted from the Matlab fminsearchbnd function
# https://uk.mathworks.com/matlabcentral/fileexchange/8277-fminsearchbnd-fminsearchcon


class BoundClass(Enum):
    UNCONSTRAINED = auto()
    LB = auto()
    UB = auto()
    BOTH = auto()
    FIXED_VAR = auto()


def get_bound_class(lb: float, ub: float) -> BoundClass:
    finite_bounds: tuple[bool, bool] = (
        bool(np.isfinite(lb)),
        bool(np.isfinite(ub)),
    )
    match finite_bounds:
        case (False, False):
            return BoundClass.UNCONSTRAINED
        case (True, False):
            return BoundClass.LB
        case (False, True):
            return BoundClass.UB
        case _:
            if lb == ub:
                return BoundClass.FIXED_VAR
            else:
                return BoundClass.BOTH


def xtransform_to_unconstrained(x0, params):
    # transform starting values into their unconstrained
    # surrogates. Check for infeasible starting guesses.
    x0u = []
    k = 0
    LB = params["LB"]
    UB = params["UB"]
    n = len(x0)
    for i in range(n):
        match params["BoundClass"][i]:
            case BoundClass.UNCONSTRAINED:
                x0u.append(x0[i])
                k += 1
            case BoundClass.LB:
                if x0[i] <= LB[i]:
                    x0u.append(0)
                else:
                    x0u.append(np.sqrt(x0[i] - LB[i]))
                k += 1
            case BoundClass.UB:
                if x0[i] >= UB[i]:
                    x0u.append(0)
                else:
                    x0u.append(np.sqrt(UB[i] - x0[i]))
                k += 1
            case BoundClass.BOTH:
                if x0[i] <= LB[i]:
                    x0u.append(-np.pi / 2)
                elif x0[i] >= UB[i]:
                    x0u.append(np.pi / 2)
                else:
                    temp = 2 * (x0[i] - LB[i]) / (UB[i] - LB[i]) - 1
                    # shift by 2*pi to avoid problems at zero in fminsearch
                    # otherwise, the initial simplex is vanishingly small
                    x0u.append(2 * np.pi + np.arcsin(np.clip(temp, -1, 1)))
                k += 1
            case BoundClass.FIXED_VAR:
                # fixed variable. drop it before fminsearch sees it.
                # k is not incremented for this variable.
                pass
    return np.array(x0u)


def xtransform_to_constrained(x0u, params):
    xtrans = []
    k = 0
    for i in range(params["n"]):
        match params["BoundClass"][i]:
            case BoundClass.UNCONSTRAINED:
                xtrans.append(x0u[k])
                k += 1
            case BoundClass.LB:
                xtrans.append(params["LB"][i] + x0u[k] ** 2)
                k += 1
            case BoundClass.UB:
                xtrans.append(params["UB"][i] - x0u[k] ** 2)
                k += 1
            case BoundClass.BOTH:
                temp = ((np.sin(x0u[k]) + 1) / 2) * (
                    params["UB"][i] - params["LB"][i]
                ) + params["LB"][i]

                xtrans.append(
                    np.maximum(
                        params["LB"][i], np.minimum(params["UB"][i], temp)
                    )
                )
                k += 1
            case BoundClass.FIXED_VAR:
                xtrans.append(params["LB"][i])
    return np.array(xtrans)


Nfeval = 1


def fminsearchbnd(
    fun, x0, LB=None, UB=None, options=None, func_args=[], *args, **kwargs
):
    def intrafun(x, params):
        xtrans = xtransform_to_constrained(x, params).reshape(params["xsize"])
        return fun(xtrans, *params["args"])

    xsize = np.atleast_1d(np.asarray(x0).shape)
    x0 = np.asarray(x0).ravel()
    n = len(x0)

    if LB is None or len(LB) == 0:
        LB = -np.inf * np.ones(n)
    else:
        LB = np.asarray(LB).ravel()

    if UB is None or len(UB) == 0:
        UB = np.inf * np.ones(n)
    else:
        UB = np.asarray(UB).ravel()

    if n != len(LB) or n != len(UB):
        raise ValueError("x0 is incompatible in size with either LB or UB.")

    if options is None:
        options = {"disp": False}

    params = {
        "args": func_args,
        "LB": LB,
        "UB": UB,
        "fun": fun,
        "n": n,
        "xsize": xsize,
        "OutputFcn": None,
    }

    params["BoundClass"] = [get_bound_class(LB[i], UB[i]) for i in range(n)]

    x0_unconstrained = xtransform_to_unconstrained(x0, params)

    if len(x0_unconstrained) == 0:
        # All variables were fixed. quit immediately
        result = OptimizeResult()
        result["x"] = x0
        result["success"] = False
        result["fun"] = fun(x0, *args)

        return result

    def callbackFn(Xi):
        f = intrafun(Xi, params)
        global Nfeval
        print(
            "{0:4d}   {1: 3.6e}   {2: 3.6e}   {3: 3.6e}   {4: 3.6e}   {5: 3.6e}   {6: 3.6e}".format(
                Nfeval, Xi[0], Xi[1], Xi[2], Xi[3], Xi[4], f
            )
        )
        Nfeval += 1

    callback = None

    if options["disp"]:
        callback = callbackFn

    result = minimize(
        intrafun,
        x0_unconstrained,
        args=(params,),
        **kwargs,
        options=options,
        method="Nelder-Mead",
        callback=callback,
    )

    x_unconstrained = result.x
    x = xtransform_to_constrained(x_unconstrained, params).reshape(xsize)

    transformed_result = result.copy()
    transformed_result["x"] = x
    return transformed_result



from enum import Enum, auto
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from .model_types import ExtrapolatedBoundaryConditions, ZeroBoundaryConditions


class BoundaryType(Enum):
    ZBC = auto()
    EBC = auto()


class QuantityType(Enum):
    REFLECTANCE = auto()
    ATTENUATION = auto()
    ATTENUATION_SLOPE = auto()


def get_model(
    boundary_condition_type: BoundaryType,
    quantity: QuantityType,
    distance_max: Optional[float],
):
    model_choice = (boundary_condition_type, quantity)
    match model_choice:
        case (BoundaryType.ZBC, QuantityType.REFLECTANCE):
            model_function = ZeroBoundaryConditions.reflectance()
        case (BoundaryType.ZBC, QuantityType.ATTENUATION):
            model_function = ZeroBoundaryConditions.attenuation()
        case (BoundaryType.ZBC, QuantityType.ATTENUATION_SLOPE):
            model_function = ZeroBoundaryConditions.attenuation_slope(
                is_long_separation=(distance_max is not None)
            )
        case (BoundaryType.EBC, QuantityType.REFLECTANCE):
            model_function = ExtrapolatedBoundaryConditions.reflectance()
        case (BoundaryType.EBC, QuantityType.ATTENUATION):
            model_function = ExtrapolatedBoundaryConditions.attenuation()
        case (BoundaryType.EBC, QuantityType.ATTENUATION_SLOPE):
            model_function = ExtrapolatedBoundaryConditions.attenuation_slope(
                is_long_separation=(distance_max is not None)
            )
        case _:
            raise ValueError(
                f"Incorrect model type passed. Passed {model_choice} but \
                    should be of type (BoundaryType, QuantityType)"
            )

    return model_function


def derivative_fit(
    param: NDArray[np.float64],
    boundary_condition_type: BoundaryType,
    quantity: QuantityType,
    slope_diff: NDArray[np.float64],
    extinction: NDArray[np.float64],
    wavelengths: NDArray[np.float64],
    distance: float,
    distance_max: Optional[float] = None,
    wave_start: float = 710.0,
    wave_end: float = 900.0,
) -> np.floating:
    """Create objective function to fit derivative

    Args:
        param (NDArray[np.float64]): First is water fraction, second is
        HHb, third is HbO2, fourth and fifth are the scattering coefficients
        from the exponential model; a and b resp., from mu_s = a*lambda^(-b)
        where lambda is in micrometers.

        boundary_condition_type (BoundaryType): Boundary type e.g. Zero or
        extrapolated

        quantity (QuantityType): Quantity to obtain. One of [reflectance,
        attenuation, attenuation slope].

        slope_diff (NDArray[np.float64]): Differential of slope. See
        calc_values.py for derivation

        extinction (NDArray[np.float64]): Extinction co-efficients matrix,
        W x 4, first column wavelength, second HHb, third HbO2, fourth the
        absorption coeff of water.

        wavelengths (NDArray[np.float64]): W x 1 array of wavelengths

        distance (float): Source-detector separation. If separation between
        detectors is negligible this is rho. If non-negligible this is the
        minimum distance from source to detector.

        distance_max (Optional[float], optional): Maximal source-detector
        distance. Not needed it distance bwteen detectors is negligible.
        Defaults to None.

        wave_start (int, optional): Start of wavelength range fitting is
        performed on. Defaults to 710.

        wave_end (int, optional): End of wavelength range fitting is performed
        on. Defaults to 900.

    Raises:
        ValueError: ValueError for unable to find unique start and end
        wavelengths

    Returns:
        np.floating: sum of least square differences
    """
    start_idx = np.argwhere(wavelengths == wave_start)
    end_idx = np.argwhere(wavelengths == wave_end)

    if (start_idx.shape != (1, 1)) or (end_idx.shape != (1, 1)):
        raise ValueError("Couldn't find unique start and end wavelengths")

    start_idx = start_idx[0][0]
    end_idx = end_idx[0][0]

    water_fraction, hhb_fraction, hbo2_fraction, a, b = param

    mu_a = water_fraction * extinction[:, 3] + np.log(10) * (
        hhb_fraction * extinction[:, 1] + hbo2_fraction * extinction[:, 2]
    )
    mu_s = a * (wavelengths * 0.001) ** (-b)

    rho = distance

    d_s, d_l = distance, distance_max

    model_function = get_model(boundary_condition_type, quantity, distance_max)
    if distance_max:
        slope_model_result = model_function(mu_s, mu_a, d_s, d_l)
    else:
        slope_model_result = model_function(mu_s, mu_a, rho)

    slope_model_diff = np.diff(slope_model_result, n=1)

    difference = (
        slope_model_diff[start_idx : end_idx + 1]
        - slope_diff[start_idx : end_idx + 1]
    )

    least_square = np.sum(difference**2)
    return least_square



import numpy as np
from numpy.typing import NDArray
from scipy.stats import linregress


def calc_attenuation_spectra(
    intensity_spectra: NDArray, ref_spectra: NDArray
) -> NDArray:
    """Calculate attenuation spectra from intensities

    Args:
        intensity_spectra (NDArray): Intensity spectra
        ref_spectra (NDArray): Re:w
        ference spectrum

    Returns:
        NDArray: Calculated attenuation spectra. Will convert 1D array to 2D
    """

    if intensity_spectra.ndim == 1:
        intensity_spectra = intensity_spectra.reshape(
            -1, len(intensity_spectra)
        )
    return np.log10(np.divide(ref_spectra, intensity_spectra))


def calc_attenuation_slope(
    attenuation_spectra: NDArray, source_detector_distances: NDArray
) -> NDArray:
    """Calculate the attenuation slope via linear regression.

    Calculate attenuation slope at each of `T` timepoints for a spectra with
    `N` wavelengths and `k` source-detector distances

    Args:
        attenuation_spectra (NDArray): Matrix of attenuation spectra of shape
        `T`x`N`x`k`
        source_detector_distances (NDArray): List of source-detector distances.
        Should be of length `k`

    Returns:
        NDArray: Matrix of attenuation slope for each timepoint. Shape `T`x`N`
    """

    k, T, N = attenuation_spectra.shape

    if len(source_detector_distances) != k:
        raise ValueError(
            f"Mismatch between numbers of distances and layers of attenuation\
                matrix.\n\
                Got {len(source_detector_distances)} and {k} respectively."
        )

    def get_slope(attenuations: NDArray, distances: NDArray):
        result = linregress(distances, attenuations)
        return result.slope  # type: ignore

    return np.apply_along_axis(
        get_slope, 0, attenuation_spectra, source_detector_distances
    )



__all__ = [
    "calc_values",
    "smooth",
    "derivative_fit",
    "get_model",
    "ZeroBoundaryConditions",
    "ExtrapolatedBoundaryConditions",
    "BoundaryType",
    "QuantityType",
    "Boundaries",
]
from .boundaries import Boundaries
from .calc_values import calc_values, smooth
from .derivative_fit import (
    BoundaryType,
    QuantityType,
    derivative_fit,
    get_model,
)
from .model_types import ExtrapolatedBoundaryConditions, ZeroBoundaryConditions



import sympy
from sympy import lambdify


# ~~~~~~~~~~~~~~ Define unresolved symbolic equations ~~~~~~~~~~~~~~~~~~~~~~~ #
def zbc_reflectance(mu_s, mu_a, rho):
    z0 = 1 / mu_s
    mueff = sympy.sqrt(3 * mu_a * mu_s)
    return (z0 * mueff * sympy.exp(-mueff * rho)) / (2 * sympy.pi * rho**2)


def zbc_attenuation(mu_s, mu_a, rho):
    z0 = 1 / mu_s
    mueff = sympy.sqrt(3 * mu_a * mu_s)
    return (
        (mueff * rho)
        + (2.0 * sympy.log(rho))  # type: ignore
        - sympy.log((z0 * mueff) / (2 * sympy.pi))
    ) / sympy.log(10)


def zbc_attenuation_slope_short_separation(mu_s, mu_a, rho):
    return (sympy.sqrt(3 * mu_s * mu_a) + 2 / rho) / sympy.log(10)


def zbc_attenuation_slope_long_separation(mu_s, mu_a, d_s, d_l):
    return (
        sympy.sqrt(3 * mu_s * mu_a)
        + 2 * ((sympy.log(d_l / d_s)) / (d_l - d_s))
    ) / sympy.log(10)


# ~~~~~~~~~~~~~ End unresolve symbolic equations ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class ZeroBoundaryConditions:
    @staticmethod
    def reflectance():
        mu_s, mu_a, rho = sympy.symbols("mu_s mu_a rho")
        return lambdify(
            [mu_s, mu_a, rho], zbc_reflectance(mu_s, mu_a, rho), "numpy"
        )

    @staticmethod
    def attenuation():
        mu_s, mu_a, rho = sympy.symbols("mu_s mu_a rho")
        return lambdify(
            [mu_s, mu_a, rho], zbc_attenuation(mu_s, mu_a, rho), "numpy"
        )

    @staticmethod
    def attenuation_slope(is_long_separation: bool = False):
        mu_s, mu_a, rho, d_s, d_l = sympy.symbols("mu_s mu_a rho d_s d_l")

        if is_long_separation:
            return lambdify(
                [mu_s, mu_a, d_s, d_l],
                zbc_attenuation_slope_long_separation(mu_s, mu_a, d_s, d_l),
                "numpy",
            )

        return lambdify(
            [mu_s, mu_a, rho],
            zbc_attenuation_slope_short_separation(mu_s, mu_a, rho),
            "numpy",
        )


# ~~~~~~~~~~~~~~ Define unresolved symbolic equations ~~~~~~~~~~~~~~~~~~~~~~~ #
def ebc_reflectance(mu_s, mu_a, rho):
    z0 = 1 / mu_s
    mueff = sympy.sqrt(3 * mu_a * mu_s)
    D = 1.0 / (3 * (mu_a + mu_s))

    zb = (
        (1 + 0.493) / (1 - 0.493) * 2 * D
    )  # from Kienle(1997),j. Opt. Soc. Am. A 14:1.
    # Valid for biological tissue.
    r1sq = (
        rho**2
    )  # r1^2 = z0^2 + rho^2 but z0^2 is negligible compared to rho^2
    r2 = (z0 + 2 * zb) ** 2 + rho**2
    return (
        1
        / (4 * sympy.pi)
        * (
            z0
            * (mueff + 1.0 / sympy.sqrt(r1sq))  # type: ignore
            * (sympy.exp(-mueff * sympy.sqrt(r1sq)) / r1sq)  # type: ignore
            + (z0 + 2 * zb)
            * (mueff + 1.0 / sympy.sqrt(r2))  # type: ignore
            * (sympy.exp(-mueff * sympy.sqrt(r2)) / r2)  # type: ignore
        )
    )


def ebc_attenuation(mu_s, mu_a, rho):
    return -1.0 * sympy.log(
        ebc_reflectance(mu_s, mu_a, rho), 10
    )  # type: ignore


def ebc_attenuation_slope_short_separation(mu_s, mu_a, rho):
    return sympy.diff(ebc_attenuation(mu_s, mu_a, rho), rho)


def ebc_attenuation_slope_long_separation(mu_s, mu_a, d_s, d_l):
    return (
        ebc_attenuation(mu_s, mu_a, d_l) - ebc_attenuation(mu_s, mu_a, d_s)
    ) / (
        d_l - d_s
    )  # type: ignore


# ~~~~~~~~~~~~~ End unresolve symbolic equations ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class ExtrapolatedBoundaryConditions:
    @staticmethod
    def reflectance():
        mu_s, mu_a, rho = sympy.symbols("mu_s mu_a rho")
        return lambdify(
            [mu_s, mu_a, rho], ebc_reflectance(mu_s, mu_a, rho), "numpy"
        )

    @staticmethod
    def attenuation():
        mu_s, mu_a, rho = sympy.symbols("mu_s mu_a rho")
        return lambdify(
            [mu_s, mu_a, rho], ebc_attenuation(mu_s, mu_a, rho), "numpy"
        )

    @staticmethod
    def attenuation_slope(is_long_separation: bool = False):
        mu_s, mu_a, rho, d_s, d_l = sympy.symbols("mu_s mu_a rho d_s d_l")

        if is_long_separation:
            return lambdify(
                [mu_s, mu_a, d_s, d_l],
                ebc_attenuation_slope_long_separation(mu_s, mu_a, d_s, d_l),
                "numpy",
            )

        return lambdify(
            [mu_s, mu_a, rho],
            ebc_attenuation_slope_short_separation(mu_s, mu_a, rho),
            "numpy",
        )



import numpy as np


def calc_dpf(mu_s: float, mu_a: float, d: float) -> float:
    """Calculate differential pathlength factor from mu_a and mu_s

    Taken from https://doi.org/10.1117/1.JBO.18.10.105004

    Args:
        mu_s (float): reduced scattering coefficient
        mu_a (float): absorption coefficient
        d (float): source-detector distance

    Returns:
        float: differential pathlength factor
    """

    return (
        0.5
        * np.sqrt(3 * mu_s / mu_a)
        * (1 - (1 / (1 + np.sqrt(d * 3 * mu_a * mu_s))))
    )



from dataclasses import dataclass

import numpy as np

boundaries = np.array(
    [
        [1, 20, 20, 1, 3],
        [0.970000000000000, 0, 0, 0, 0],
        [1, 40, 40, 2, 4],
    ]
)


@dataclass
class Boundaries:
    boundaries = boundaries
    columns = ["water_frac", "HHb", "HbO2", "a", "b"]



from typing import Optional

import numpy as np
from numpy.typing import NDArray

from .derivative_fit import (
    BoundaryType,
    QuantityType,
    derivative_fit,
    get_model,
)
from .fminsearchbnd import fminsearchbnd


def smooth(a: NDArray[np.float64], span: int) -> NDArray:
    """MATLAB Smooth function clone

    Args:
        a (np.ndarray): Numpy array of data to smooth
        span (int): Size of smoothing window. Must be an odd number

    Returns:
        np.ndarray: Smoothed data
    """

    out0 = np.convolve(a, np.ones(span, dtype=int), "valid") / span
    r = np.arange(1, span - 1, 2)
    start = np.cumsum(a[: span - 1])[::2] / r
    stop = (np.cumsum(a[:-span:-1])[::2] / r)[::-1]
    return np.concatenate((start, out0, stop))


def calc_values(
    slope: np.ndarray,
    extinction: np.ndarray,
    wavelengths: np.ndarray,
    boundaries: np.ndarray,
    boundary_condition_type: BoundaryType,
    distance: float,
    distance_max: Optional[float] = None,
):
    """Calculate parameters by fitting attenuation slope

    For more detail see Chapter 8 of "A Novel Approach to Monitor Tissue Oxygen
    Saturation with Broadband Near-Infrared Spectroscopy" - Zuzanna Kovacsova

    Args:
        slope (np.ndarray): Attenuation slope
        extinction (np.ndarray): Matrix of extinction co-efficients for each species
        and wavelength
        wavelengths (np.ndarray): Wavelengths of light used
        boundaries (np.ndarray): Boundaries for parameters. First row is start,
        second is lower bound, third is upper bound
        boundary_condition_type (BoundaryType): Zero or Extrapolated boundary condition
        distance (float): Distance between source and detector. If one distance used
        this is it. If maximal distance used, this is the minimal.
        distance_max (Optional[float], optional): Optional maximum distance.
        Defaults to None.

    Raises:
        RuntimeError: Error if fails to obtain co-efficients.

    Returns:
        tuple: Tuple of stO2, coefficients, residual, residual_norm,
        sum_residual, score
    """
    start = boundaries[0]
    LB = boundaries[1]
    UB = boundaries[2]

    slope_1stdiff = np.diff(smooth(slope, 5))
    # Set wavelength range for the fitting
    wave_start = 710
    wave_end = 900

    # Setting fitting options
    options = {
        "disp": False,
        "maxiter": 200000,
        "maxfev": 200000,
        "xatol": 1e-10,
        "fatol": 1e-10,
    }

    result = fminsearchbnd(
        derivative_fit,
        x0=start,
        LB=LB,
        UB=UB,
        func_args=(
            boundary_condition_type,
            QuantityType.ATTENUATION_SLOPE,
            slope_1stdiff,
            extinction,
            wavelengths,
            distance,
            distance_max,
            wave_start,
            wave_end,
        ),
        options=options,
        tol=1e-10,
    )

    if result["success"]:
        coefficients = result["x"]
    else:
        raise RuntimeError("Failed to solve for coefficients.")

    mua = coefficients[0] * extinction[:, 3] + np.log(10) * (
        coefficients[1] * extinction[:, 1] + coefficients[2] * extinction[:, 2]
    )
    mus = coefficients[3] * (wavelengths * 0.001) ** (-coefficients[4])

    model_function = get_model(
        boundary_condition_type, QuantityType.ATTENUATION_SLOPE, distance_max
    )
    if distance_max:
        model_result = model_function(distance_max, distance, mua, mus)
    else:
        model_result = model_function(mua, mus, distance)

    model_1stdiff = np.diff(model_result, n=1)

    stO2 = coefficients[2] / (coefficients[1] + coefficients[2]) * 100

    # Determine the score of the fit

    residual = (model_1stdiff - slope_1stdiff) ** 2
    sum_residual = np.sum(residual)

    # Indicies of values to determine range between
    index_710 = np.where(wavelengths == 710)[0][0]
    index_900 = np.where(wavelengths == 900)[0][0]

    # Indices of HHb wavelengths
    index_HHb = np.arange(
        np.where(wavelengths == 750)[0][0],
        np.where(wavelengths == 770)[0][0] + 1,
    )

    # Indices of H2O wavelengths
    index_water = np.arange(
        np.where(wavelengths == 825)[0][0],
        np.where(wavelengths == 840)[0][0] + 1,
    )

    residual_norm = (
        model_1stdiff / np.max(model_1stdiff)
        - slope_1stdiff / np.max(model_1stdiff)
    ) ** 2

    sum_hhb_residuals = np.sum(residual_norm[index_HHb])
    sum_water_residuals = np.sum(residual_norm[index_water])

    model_range = np.max(
        model_1stdiff[index_710:index_900]
        / np.max(model_1stdiff[index_710:index_900])
    ) - np.min(
        model_1stdiff[index_710:index_900]
        / np.max(model_1stdiff[index_710:index_900])
    )

    score = sum_hhb_residuals * sum_water_residuals / model_range

    return stO2, coefficients, residual, residual_norm, sum_residual, score



from .attenuation import calc_attenuation_slope, calc_attenuation_spectra
from .dpf import calc_dpf
from .extinction_coefficients import ExtinctionCoefficients

__all__ = [
    "calc_dpf",
    "calc_attenuation_spectra",
    "calc_attenuation_slope",
    "ExtinctionCoefficients",
]



from typing import List
import pandas as pd
import os
import numpy as np


class DefaultValues:
    def __init__(
        self,
        csv_file: str = os.path.dirname(__file__) + "/defaults.csv",
        species: List[str] = ["HbO2", "HHb", "CCO"],
    ) -> None:
        df: pd.DataFrame = pd.read_csv(
            csv_file,
            sep=",",
            engine="pyarrow",
            index_col="wavelength",
        )

        self.extinction_coefficients = df[species].values
        self.wavelength_dependency = np.array(df["wl_dep"].values)
        self.spectra_wavelengths = np.array(df.index.values)
        self.spectra_wavelengths = df.index.values
